#include <mpi.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <chrono>
#include <sstream>
using namespace std;

class Component
{
private:
    int rank, size;
    int component, nComponents, componentSize, componentRank, componentProblemSize, componentRankZero;
    vector<int> componentsRankZero;
    int nts, problemSize;
    double tsPerSecond;

    vector<double> A, B, C;

    void print(const string &str)
    {
        cout << "COMP" << component << ":RANK" << rank << ": " << str << endl;
    }

    void loadInidata()
    {
        ifstream modelIni("inidata/0");
        modelIni >> nComponents;
        modelIni >> nts;

        componentsRankZero = vector<int>(nComponents);

        int currComponent, currComponentSize;
        int definedComponents = 0;
        int nproc = 0;
        while (modelIni >> currComponent >> currComponentSize)
        {
            if (rank >= nproc)
            {
                component = currComponent;
                componentSize = currComponentSize;
                componentRank = rank - nproc;
                componentRankZero = nproc;
            }
            componentsRankZero[definedComponents] = nproc;
            nproc += currComponentSize;
            ++definedComponents;
        }

        modelIni.close();

        if (rank == 0)
        {
            if (nproc != size)
            {
                cerr << "[ERROR] Total number of nproc does not match the sum of the component nproc" << endl;
                cerr << "[ERROR] Check inidata/0 config file" << endl;
                MPI_Abort(MPI_COMM_WORLD, 1);
            }
            else if (definedComponents != nComponents)
            {
                cerr << "[ERROR] Number of components does not match the number of defined components" << endl;
                cerr << "[ERROR] Check inidata/0 config file" << endl;
                MPI_Abort(MPI_COMM_WORLD, 1);
            }
        }

        ifstream compIni("inidata/" + to_string(component));
        compIni >> problemSize;
        compIni.close();

        componentProblemSize = problemSize / componentSize;
    }

    void perfReport()
    {
        // Component rank 0 reports for its component
        if (rank == componentRankZero)
        {
            for (int i = 1; i < componentSize; ++i)
            {
                double tsPerSecondRecv;
                MPI_Recv(&tsPerSecondRecv, 1, MPI_DOUBLE, i + componentRankZero, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                tsPerSecond += tsPerSecondRecv;
            }
            tsPerSecond /= componentSize;

            // Rank 0 reports for all compoenents
            if (rank == 0)
            {
                double currentSypd = tsPerSecond * 2700 / 365;

                double modelTsPerSecond = tsPerSecond;
                double modelSypd = currentSypd;

                ofstream perfReportF("test_info.csv");
                perfReportF << "component,ts_per_second,sypd" << endl;
                perfReportF << "COMP1," << tsPerSecond << "," << currentSypd << endl;

                for (int i = 1; i < nComponents; ++i)
                {
                    double currentTsPerSecond;
                    print("RECV " + to_string(componentsRankZero[i]));
                    MPI_Recv(&currentTsPerSecond, 1, MPI_DOUBLE, componentsRankZero[i], 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                    currentSypd = currentTsPerSecond * 2700 / 365;
                    modelTsPerSecond = min(modelTsPerSecond, currentTsPerSecond);
                    modelSypd = min(modelSypd, currentSypd);
                    perfReportF << "COMP" << i + 1 << "," << currentTsPerSecond << "," << currentSypd << endl;
                }
                perfReportF << "model" << "," << modelTsPerSecond << "," << modelSypd << endl;
                perfReportF.close();
            }
            else
            {
                MPI_Request request;
                MPI_Isend(&tsPerSecond, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &request);
            }
        }
        else
        {
            MPI_Request request;
            MPI_Isend(&tsPerSecond, 1, MPI_DOUBLE, componentRankZero, 0, MPI_COMM_WORLD, &request);
        }
    }

public:
    Component()
    {
        MPI_Comm_rank(MPI_COMM_WORLD, &rank);
        MPI_Comm_size(MPI_COMM_WORLD, &size);

        loadInidata();

        A = vector<double>(componentProblemSize * componentProblemSize);
        B = vector<double>(componentProblemSize * componentProblemSize);
        C = vector<double>(componentProblemSize * componentProblemSize);

        for (int i = 0; i < componentProblemSize; ++i)
        {
            for (int j = 0; j < componentProblemSize; ++j)
            {
                A[i * componentProblemSize + j] = i * componentProblemSize + j + component;
                B[i * componentProblemSize + j] = i * componentProblemSize + j + component;
            }
        }
    }

    void run()
    {
        tsPerSecond = 0;
        for (int t = 0; t < nts; ++t)
        {
            auto start = chrono::steady_clock::now();
            for (int i = 0; i < componentProblemSize; ++i)
                for (int j = 0; j < componentProblemSize; ++j)
                    C[i * componentProblemSize + j] = A[i * componentProblemSize + j] + B[i * componentProblemSize + j];

            int receiver = (componentRank + 1) % componentSize + componentRankZero;
            int sender = (componentRank - 1) % componentSize;
            if (sender < 0)
                sender += componentSize;
            sender += componentRankZero;

            MPI_Request request;
            cout << "COMP" << component << ":RANK" << rank << ": SEND " << receiver << endl;
            MPI_Isend(&C[0], componentProblemSize * componentProblemSize, MPI_DOUBLE, receiver, 0, MPI_COMM_WORLD, &request);
            cout << "COMP" << component << ":RANK" << rank << ": RECV " << sender << endl;
            MPI_Recv(&B[0], componentProblemSize * componentProblemSize, MPI_DOUBLE, sender, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

            auto end = chrono::steady_clock::now();
            chrono::duration<double> elapsed_seconds = end - start;
            tsPerSecond += 1.0 / elapsed_seconds.count();

            MPI_Barrier(MPI_COMM_WORLD);

            auto endWaiting = chrono::steady_clock::now();
            chrono::duration<double> elapsedWaiting = endWaiting - end;

            if (rank == 0)
                cout << "Timestep: " << t << endl;
        }
        tsPerSecond /= nts;

        perfReport();
    }

    ~Component()
    {
    }
};

int main(int argc, char **argv)
{
    MPI_Init(&argc, &argv);

    Component *component = new Component();
    component->run();

    MPI_Finalize();
    return 0;
}